/*
  This file is part of GerAstro.

  GerAstro is free software: you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  GerAstro is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  GerAstro.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "cubicspline.h"


namespace GerPh {

  CubicSpline::CubicSpline(
    double aA0, double aA1, double aA2, double aA3,
    double aArgBegin, double aArgRange
  ) {
    setInterval(aArgBegin, aArgRange);
    setCoefficients(aA0, aA1, aA2, aA3);
  } // CubicSpline::CubicSpline

  CubicSpline::CubicSpline(
    double aValBegin, double aDrvBegin, double aValEnd, double aDrvEnd,
    int aDrvOrder, double aArgBegin, double aArgRange
  ) {
    setInterval(aArgBegin, aArgRange);
    setBoundaries(aValBegin, aDrvBegin, aValEnd, aDrvEnd, aDrvOrder);
  } // CubicSpline::CubicSpline

  double CubicSpline::getArgBegin() {
    return fArgBegin;
  } // CubicSpline::getArgBegin

  double CubicSpline::getArgEnd() {
    return fArgBegin + fArgRange;
  } // CubicSpline::getArgEnd

  double CubicSpline::getArgRange() {
    return fArgRange;
  } // CubicSpline::getRange

  double CubicSpline::getDerivative(double aArg) {
    double p = (aArg - fArgBegin) * fRezArgRange;
    return (fA1 + p * (2.0 * fA2 + p * 3.0 * fA3)) * fRezArgRange;
  } // CubicSpline::getDerivative

  double CubicSpline::getDrvBegin() {
    return fA1 * fRezArgRange;
  } // CubicSpline::getDrvBegin

  double CubicSpline::getDrvEnd() {
    return (fA1 + 2.0 * fA2 + 3.0 * fA3) * fRezArgRange;
  } // CubicSpline::getDrvEnd

  double CubicSpline::getValBegin() {
    return fA0;
  } // CubicSpline::getValBegin

  double CubicSpline::getValEnd() {
    return fA0 + fA1 + fA2 + fA3;
  } // CubicSpline::getValEnd

  double CubicSpline::getValue(double aArg) {
    double p = (aArg - fArgBegin) * fRezArgRange;
    return fA0 + p * (fA1 + p * (fA2 + p * fA3));
  } // CubicSpline::getValue

  double CubicSpline::getValueLinear(double aArg) {
    double p = (aArg - fArgBegin) * fRezArgRange;
    return fA0 + (fA1 + fA2 + fA3) * p;
  } // CubicSpline::getValueLinear

  void CubicSpline::setBoundaries(
    double aValBegin, double aDrvBegin, double aValEnd, double aDrvEnd,
    int aDrvOrder
  ) {
    //!!! TODO: implement aDrvOrder other than 1!
    fA0 = aValBegin;
    fA1 = aDrvBegin * fArgRange;
    fA3 = (aDrvBegin + aDrvEnd) * fArgRange + 2.0 * (aValBegin - aValEnd);
    fA2 = aValEnd - (fA3 + fA1 + fA0);
  } // CubicSpline::setBoundaries

  void CubicSpline::setCoefficients(
    double aA0, double aA1, double aA2, double aA3
  ) {
    fA0 = aA0;
    fA1 = aA1;
    fA2 = aA2;
    fA3 = aA3;
  } // CubicSpline::setCoefficients

  void CubicSpline::setInterval(double aArgBegin, double aArgRange) {
    fArgBegin = aArgBegin;
    fArgRange = aArgRange;
    fRezArgRange = 1.0 / fArgRange;
  } // CubicSpline::setInterval

} // GerPh
