/*
  This file is part of GerAstro.

  GerAstro is free software: you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  GerAstro is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  GerAstro.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "defaultsplinefunction.h"

#include "gaussian.h"


namespace GerPh {

  DefaultSplineFunction::DefaultSplineFunction(
    int aCount, double *aArgs, double *aValues,
    double aDrvBegin, double aDrvEnd, bool aDamped
  ) {
    if (aDamped) {
      initDamped(aCount, aArgs, aValues, aDrvBegin, aDrvEnd);
    } else {
      initialize(aCount, aArgs, aValues, aDrvBegin, aDrvEnd);
    } // if
  } // DefaultSplineFunction::DefaultSplineFunction

  DefaultSplineFunction::DefaultSplineFunction(
    int aCount, double *aValues, double aDrvBegin, double aDrvEnd, bool aDamped
  ) {
    double args[aCount];
    for (int i = 0; i < aCount; i++) {
      args[i] = i;
    } // for
    if (aDamped) {
      initDamped(aCount, args, aValues, aDrvBegin, aDrvEnd);
    } else {
      initialize(aCount, args, aValues, aDrvBegin, aDrvEnd);
    } // if
  } // DefaultSplineFunction::DefaultSplineFunction

  DefaultSplineFunction::~DefaultSplineFunction() {
    while (fSplines -> getCount() > 0) {
      CubicSpline *spline = fSplines -> getAt(0);
      fSplines -> removeAt(0);
      delete spline;
    } // while
    delete fSplines;
  } // DefaultSplineFunction::~DefaultSplineFunction

  int DefaultSplineFunction::cmpByStart(
    CubicSpline *aSpline1, CubicSpline *aSpline2
  ) {
    double start1 = aSpline1 -> getArgBegin();
    double start2 = aSpline2 -> getArgBegin();
    if (start1 < start2) {
      return -1;
    } else if (start1 > start2) {
      return +1;
    } // if
    return 0;
  } // DefaultSplineFunction::cmpByStart

  int DefaultSplineFunction::getCountSplines() {
    return fSplines -> getCount();
  } // DefaultSplineFunction::getCountSplines

  CubicSpline* DefaultSplineFunction::getSpline(double aArg) {
    //!!! TODO: Change to log search
    int n = fSplines -> getCount();
    int i = 0;
    bool found = false;
    CubicSpline *spline = NULL;
    while (i < n && !found) {
      spline = fSplines -> getAt(i);
      double x0 = spline -> getArgBegin();
      double x1 = x0 + spline -> getArgRange();
      found = aArg >= x0 && aArg <= x1;
      if (!found) {
        i++;
      } // if
    } // while
    if (found) {
      return spline;
    } // if
    return NULL;
  } // DefaultSplineFunction::getSpline

  CubicSpline* DefaultSplineFunction::getSplineAt(int aIndex) {
    return fSplines -> getAt(aIndex);
  } // DefaultSplineFunction::getSplineAt

  double DefaultSplineFunction::getValue(double aArg) {
    CubicSpline *spline = getSpline(aArg);
    if (!spline) {
      //!!! TODO: throw an exception?
      return 0.0;
    } // if
    return spline -> getValue(aArg);
  } // DefaultSplineFunction::getValue

  double DefaultSplineFunction::getValueLinear(double aArg) {
    CubicSpline *spline = getSpline(aArg);
    if (!spline) {
      //!!! TODO: throw an exception?
      return 0.0;
    } // if
    return spline -> getValueLinear(aArg);
  } // DefaultSplineFunction::getValueLinear

  void DefaultSplineFunction::initialize(
    int aCount, double *aArgs, double *aValues,
    double aDrvBegin, double aDrvEnd
  ) {
    fSplines = new GerQ::SortedList<CubicSpline*>(cmpByStart);
    if (aCount < 2) {
      return;
    } // if
    double ranges[aCount - 1];
    for (int i = 0; i < aCount - 1; i++) {
      ranges[i] = aArgs[i + 1] - aArgs[i];
    } // for
    int order = (aCount - 1) * 4;
    Gaussian gaussian(order);
    gaussian.newEquation() ->
      setAt(0, 1.0) ->
      setAt(order, aValues[0]);
    gaussian.newEquation() ->
      setAt(0, 1.0) -> setAt(1, 1.0) -> setAt(2, 1.0) -> setAt(3, 1.0) ->
      setAt(order, aValues[1]);
    double h;
    double sqr_h;
    double hNext = ranges[0];
    double sqr_hNext = hNext * hNext;
    gaussian.newEquation() ->
      setAt(2, 2.0) ->
      setAt(order, sqr_hNext * aDrvBegin);
    for (int i = 1; i < aCount - 1; i++) {
      int offset = i * 4;
      h = hNext;
      sqr_h = sqr_hNext;
      hNext = ranges[i];
      sqr_hNext = hNext * hNext;
      gaussian.newEquation() ->
        setAt(offset - 3, hNext) ->
	setAt(offset - 2, 2.0 * hNext) ->
	setAt(offset - 1, 3.0 * hNext) ->
	setAt(offset + 1, -h);
      gaussian.newEquation() ->
        setAt(offset - 2, sqr_hNext) ->
	setAt(offset - 1, 3.0 * sqr_hNext) ->
	setAt(offset + 1, -sqr_h);
      gaussian.newEquation() ->
        setAt(offset, 1.0) ->
	setAt(order, aValues[i]);
      gaussian.newEquation() ->
        setAt(offset, 1.0) ->
        setAt(offset + 1, 1.0) ->
        setAt(offset + 2, 1.0) ->
        setAt(offset + 3, 1.0) ->
	setAt(order, aValues[i + 1]);
    } // for
    h = ranges[aCount - 1];
    gaussian.newEquation() ->
      setAt(order - 2, 2.0) -> setAt(order - 1, 6.0) ->
      setAt(order, h * h * aDrvEnd);
    gaussian.solve();
    for (int i = 0; i < aCount - 1; i++) {
      double offset = i * 4;
      double a0 = gaussian.getValueAt(offset);
      double a1 = gaussian.getValueAt(offset + 1);
      double a2 = gaussian.getValueAt(offset + 2);
      double a3 = gaussian.getValueAt(offset + 3);
      fSplines -> insert(
        new CubicSpline(a0, a1, a2, a3, aArgs[i], ranges[i]));
    } // for
  } // DefaultSplineFunction::initialize

  void DefaultSplineFunction::initDamped(
    int aCount, double *aArgs, double *aValues,
    double aDrvBegin, double aDrvEnd
  ) {
    fSplines = new GerQ::SortedList<CubicSpline*>(cmpByStart);
    if (aCount < 2) {
      return;
    } // if
    double x = aArgs[0];
    double xNext = aArgs[1];
    double f = aValues[0];
    double df = aDrvBegin;
    double fNext = aValues[1];
    double dfNext = aDrvEnd;
    double delta = 2.0;
    if (aCount > 2) {
      delta = aArgs[2] - x;
      dfNext = (aValues[2] - f) / delta;
    } // if
    fSplines -> insert(new CubicSpline(f, df, fNext, dfNext, 1, x, xNext - x));
    int i = 1;
    while (i < aCount - 2) {
      x = xNext;
      xNext = aArgs[i + 1];
      delta = aArgs[i + 2] - x;
      f = fNext;
      fNext = aValues[i + 1];
      df = dfNext;
      dfNext = (aValues[i + 2] - f) / delta;
      fSplines -> insert(
        new CubicSpline(f, df, fNext, dfNext, 1, x, xNext - x));
      i++;
    } // while
    if (i < aCount - 1) {
      x = xNext;
      xNext = aArgs[aCount - 1];
      f = fNext;
      fNext = aValues[aCount - 1];
      df = dfNext;
      dfNext = aDrvEnd;
      fSplines -> insert(
        new CubicSpline(f, df, fNext, dfNext, 1, x, xNext - x));
    } // if
  } // DefaultSplineFunction::initDamped

} // GerPh
