/*
  This file is part of GerPhys.

  GerPhys is free software: you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  GerPhys is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  GerPhys.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "floatmap.h"

#include "projection.h"
#include <cmath>


namespace GerPh {

  const int FloatMap::ALIGN_CENTER	= 0;
  const int FloatMap::ALIGN_LEFT	= -1;
  const int FloatMap::ALIGN_RIGHT	= +1;
  const int FloatMap::ALIGN_TOP		= +1;
  const int FloatMap::ALIGN_BOTTOM	= -1;

  const double FloatMap::NO_FIX_RATIO = -1.0;

  FloatMap::FloatMap(QPaintDevice *aDevice) {
    fDevice = aDevice;
    fProjection = NULL;
    fDrawing = NULL;
    fDrawArgs = NULL;
    fXMin = 0.0;
    fXMax = 1.0;
    fYMin = 0.0;
    fYMax = 1.0;
    fActualXMin = 0.0;
    fActualYMax = 1.0;
    fDeltaX_W = 1.0;
    fW_DeltaX = 1.0;
    fDeltaY_H = 1.0;
    fH_DeltaY = 1.0;
    fMarginLeft = 0;
    fMarginRight = 0;
    fMarginTop = 0;
    fMarginBottom = 0;
    setFixRatioXY(NO_FIX_RATIO);
  } // FloatMap::FloatMap

  FloatMap::~FloatMap() {
    if (fProjection) {
      Projection *projection = fProjection;
      fProjection = NULL;
      projection -> setMap(NULL);
    } // if
  } // FloatMap::~FloatMap

  void FloatMap::drawEllipse(
    double aX, double aY, double aRX, double aRY, QPainter *aPainter
  ) {
    int x = getXView(aX);
    int y = getYView(aY);
    int rx = getDeltaXView(aRX);
    int ry = getDeltaYView(aRY);
    aPainter -> drawEllipse(QPoint(x, y), rx, ry);
  } // FloatMap::drawEllipse

  void FloatMap::drawEllipseSym(
    double aX, double aY, int aRX, int aRY, QPainter *aPainter
  ) {
    int x = getXView(aX);
    int y = getYView(aY);
    aPainter -> drawEllipse(QPoint(x, y), aRX, aRY);
  } // FloatMap::drawEllipseSym

  void FloatMap::drawLine(
    double aXFrom, double aYFrom, double aXTo, double aYTo, QPainter *aPainter
  ) {
    int xFrom = getXView(aXFrom);
    int yFrom = getYView(aYFrom);
    int xTo = getXView(aXTo);
    int yTo = getYView(aYTo);
    aPainter -> drawLine(xFrom, yFrom, xTo, yTo);
  } // FloatMap::drawLine

  void FloatMap::drawPoint(double aX, double aY, QPainter *aPainter) {
    int x = getXView(aX);
    int y = getYView(aY);
    aPainter -> drawPoint(x, y);
  } // FloatMap::drawPoint

  void FloatMap::drawSpot(
    double aX, double aY, double aR, QPainter *aPainter, double aStart
  ) {
    drawSpotSym(aX, aY, getDeltaXView(aR), aPainter, aStart);
  } // FloatMap::drawSpot

  void FloatMap::drawSpotSym(
    double aX, double aY, int aR, QPainter *aPainter, double aStart
  ) {
    int x = getXView(aX);
    int y = getYView(aY);
    drawSpotSymView(x, y, aR, aPainter, aStart);
  } // FloatMap::drawSpotSym

  void FloatMap::drawSpotSymView(
    int aX, int aY, int aR, QPainter *aPainter, double aStart
  ) {
    QBrush oldBrush = aPainter -> brush();
    QPen oldPen = aPainter -> pen();
    QColor color = oldBrush.color();
    QRadialGradient gradient(aX, aY, aR);
    gradient.setColorAt(aStart, color);
    gradient.setColorAt(1.0, QColor(0, 0, 0, 0));
    QBrush newBrush = QBrush(gradient);
    newBrush.setStyle(Qt::RadialGradientPattern);
    aPainter -> setBrush(newBrush);
    aPainter -> setPen(Qt::NoPen);
    aPainter -> drawEllipse(QPoint(aX, aY), aR, aR);
    aPainter -> setBrush(oldBrush);
    aPainter -> setPen(oldPen);
  } // FloatMap::drawSpotSymView

  void FloatMap::drawText(
    double aX, double aY, const QString &aText, QPainter *aPainter,
    int aHAlign, int aVAlign, int aHOffset, int aVOffset
  ) {
    if (!fDevice) {
      return;
    } // if
    QRect geometry(0, 0, fDevice -> width(), fDevice -> height());
    QRect rect = aPainter -> boundingRect(geometry, Qt::AlignCenter, aText);
    int w = rect.width();
    int h = rect.height();
    int x = getXView(aX) + aHOffset;
    int y = getYView(aY) - aVOffset;
    switch (aHAlign) {
      case ALIGN_CENTER: x -= w / 2; break;
      case ALIGN_RIGHT:  x -= w; break;
    } // switch
    switch(aVAlign) {
      case ALIGN_CENTER: y -= h / 2; break;
      case ALIGN_BOTTOM: y -= h; break;
    } // switch
    rect.setRect(x, y, w, h);
    aPainter -> drawText(rect, Qt::AlignCenter, aText);
  } // FloatMap::drawText

  void FloatMap::fillAll(QBrush *aBrush, QPainter *aPainter) {
    if (!fDevice) {
      return;
    } // if
    aPainter -> fillRect(0, 0, fDevice -> width(), fDevice -> height(),
      *aBrush);
  } // FloatMap::fillAll

  void FloatMap::fillRect(
    double aX, double aY, double aDeltaX, double aDeltaY, QBrush *aBrush,
    QPainter *aPainter
  ) {
    int x = getXView(aX);
    int y = getYView(aY);
    int w = getDeltaXView(aDeltaX);
    int h = getDeltaYView(aDeltaY);
    aPainter -> fillRect(x, y, w, h, *aBrush);
  } // FloatMap::fillRect

  double FloatMap::getActualXMax() {
    return 2.0 * getXMiddle() - fActualXMin;
  } // FloatMap::getActualXMax

  double FloatMap::getActualXMin() {
    return fActualXMin;
  } // FloatMap::getActualXMin

  double FloatMap::getActualYMax() {
    return fActualYMax;
  } // FloatMap::getActualYMax

  double FloatMap::getActualYMin() {
    return 2.0 * getYMiddle() - fActualYMax;
  } // FloatMap::getActualYMin

  double FloatMap::getDeltaXModel(int aDeltaX) {
    return aDeltaX * fDeltaX_W;
  } // FloatMap::getDeltaXModel

  int FloatMap::getDeltaXView(double aDeltaX) {
    return aDeltaX * fW_DeltaX;
  } // FloatMap::getDeltaXView

  double FloatMap::getDeltaYModel(int aDeltaY) {
    return aDeltaY * fDeltaY_H;
  } // FloatMap::getDeltaYModel

  int FloatMap::getDeltaYView(double aDeltaY) {
    return aDeltaY * fH_DeltaY;
  } // FloatMap::getDeltaYView

  QPaintDevice* FloatMap::getDevice() {
    return fDevice;
  } // FloatMap::getDevice

  double FloatMap::getDiagonal() {
    double dx = fXMax - fXMin;
    double dy = fYMax - fYMin;
    return sqrt(dx * dx + dy * dy);
  } // FloatMap::getDiagonal

  void* FloatMap::getDrawArgs() {
    if (fProjection && !fDrawArgs) {
      fDrawArgs = fProjection -> getDrawArgs();
    } // if
    return fDrawArgs;
  } // FloatMap::getDrawArgs

  void (*FloatMap::getDrawing())(QPainter*, void*) {
    if (fProjection && !fDrawing) {
      fDrawing = fProjection -> getDrawing();
    } // if
    return fDrawing;
  } // FloatMap::getDrawing

  double FloatMap::getFixRatioXY() {
    if (fFixRatioXY > 0.0) {
      return fFixRatioXY;
    } // if
    return NO_FIX_RATIO;
  } // FloatMap::getFixRatioXY

  int FloatMap::getMarginBottom() {
    return fMarginBottom;
  } // FloatMap::getMarginBottom

  int FloatMap::getMarginLeft() {
    return fMarginLeft;
  } // FloatMap::getMarginLeft

  int FloatMap::getMarginRight() {
    return fMarginRight;
  } // FloatMap::getMarginRight

  int FloatMap::getMarginTop() {
    return fMarginTop;
  } // FloatMap::getMarginTop

  Projection* FloatMap::getProjection() {
    return fProjection;
  } // FloatMap::getProjection

  void FloatMap::getTextSize(
    const QString &aText, QFont *aFont, QPainter *aPainter,
    int &aWidth, int &aHeight
  ) {
    /*
      aPainter argument is a leftover of a former implementation, where
      QPainter.boundingRect() was used to determine the drawing size of the
      string. It is no longer needed, but kept for the sake of downwards
      compatibility.
    */
    QFontMetrics metrics(*aFont);
    QRect rect = metrics.boundingRect(aText);
    aWidth = rect.width();
    aHeight = rect.height();
  } // FloatMap::getTextSize

  double FloatMap::getXMax() {
    return fXMax;
  } // FloatMap::getXMax

  double FloatMap::getXMiddle() {
    return 0.5 * (fXMin + fXMax);
  } // FloatMap::getXMiddle

  double FloatMap::getXMin() {
    return fXMin;
  } // FloatMap::getXMin

  double FloatMap::getXModel(int aXView) {
    return fActualXMin + fDeltaX_W * (aXView - fMarginLeft);
  } // FloatMap::getXModel

  int FloatMap::getXView(double aXModel) {
    return (aXModel - fActualXMin) * fW_DeltaX + fMarginLeft;
  } // FloatMap::getXView

  double FloatMap::getYMax() {
    return fYMax;
  } // FloatMap::getYMax

  double FloatMap::getYMiddle() {
    return 0.5 * (fYMin + fYMax);
  } // FloatMap::getYMiddle

  double FloatMap::getYMin() {
    return fYMin;
  } // FloatMap::getYMin

  double FloatMap::getYModel(int aYView) {
    return fActualYMax - fDeltaY_H * (aYView - fMarginTop);
  } // FloatMap::getYModel

  int FloatMap::getYView(double aYModel) {
    return (fActualYMax - aYModel) * fH_DeltaY + fMarginTop;
  } // FloatMap::getYView

  bool FloatMap::hasFixRatioXY() {
    return fFixRatioXY > 0.0;
  } // FloatMap::hasFixRatioXY

  void FloatMap::rescale() {
    if (!fDevice) {
      return;
    } // if
    int w = fDevice -> width() - (fMarginLeft + fMarginRight);
    int h = fDevice -> height() - (fMarginTop + fMarginBottom);
    if (w == 0 || h == 0) {
      return;
    } // if
    fDeltaX_W = (fXMax - fXMin) / w;
    fDeltaY_H = (fYMax - fYMin) / h;
    if (hasFixRatioXY()) {
      if (fDeltaY_H < fFixRatioXY * fDeltaX_W) {
        fDeltaY_H = fFixRatioXY * fDeltaX_W;
	fActualXMin = fXMin;
	fActualYMax = getYMiddle() + 0.5 * h * fDeltaY_H;
      } else {
        fDeltaX_W = fDeltaY_H / fFixRatioXY;
	fActualXMin = getXMiddle() - 0.5 * w * fDeltaX_W;
	fActualYMax = fYMax;
      } // if
    } else {
      fActualXMin = fXMin;
      fActualYMax = fYMax;
    } // if
    fW_DeltaX = 1.0 / fDeltaX_W;
    fH_DeltaY = 1.0 / fDeltaY_H;
    if (fProjection) {
      fProjection -> rescale();
    } // if
  } // FloatMap::rescale

  void FloatMap::setDevice(QPaintDevice *aValue) {
    fDevice = aValue;
    rescale();
  } // FloatMap::setDevice

  void FloatMap::setDrawArgs(void *aValue) {
    fDrawArgs = aValue;
  } // FloatMap::setDrawArgs

  void FloatMap::setDrawing(void (*aValue)(QPainter*, void*)) {
    fDrawing = aValue;
  } // FloatMap::setDrawing

  void FloatMap::setFixRatioXY(double aValue) {
    fFixRatioXY = aValue;
    rescale();
  } // FloatMap::setFixRatioXY

  void FloatMap::setMarginBottom(int aValue) {
    fMarginBottom = aValue;
    rescale();
  } // FloatMap::setMarginBottom

  void FloatMap::setMarginLeft(int aValue) {
    fMarginLeft = aValue;
    rescale();
  } // FloatMap::setMarginLeft

  void FloatMap::setMarginRight(int aValue) {
    fMarginRight = aValue;
    rescale();
  } // FloatMap::setMarginRight

  void FloatMap::setMarginTop(int aValue) {
    fMarginTop = aValue;
    rescale();
  } // FloatMap::setMarginTop

  void FloatMap::setProjection(Projection *aValue, double aRatioXY) {
    if (fProjection == aValue) {
      setFixRatioXY(aRatioXY);
      return;
    } // if
    if (fProjection) {
      fProjection -> setMap(NULL);
    } // if
    fProjection = aValue;
    setFixRatioXY(aRatioXY);
    if (fProjection) {
      fProjection -> setMap(this);
    } // if
  } // FloatMap::setProjection

  void FloatMap::setXBorders(double aXMin, double aXMax) {
    if (aXMin > aXMax || std::isnan(1.0 / (aXMax - aXMin))) {
      return; //!!! TODO: throw an exception?
    } // if
    fXMin = aXMin;
    fXMax = aXMax;
    rescale();
  } // FloatMap::setXBorders

  void FloatMap::setYBorders(double aYMin, double aYMax) {
    if (aYMin > aYMax || std::isnan(1.0 / (aYMax - aYMin))) {
      return; //!!! TODO: throw an exception?
    } // if
    fYMin = aYMin;
    fYMax = aYMax;
    rescale();
  } // FloatMap::setYBorders

} // GerPh
