/*
  This file is part of GerPhys.

  GerPhys is free software: you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  GerPhys is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  GerPhys.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "gaussian.h"

#include <cmath>


namespace GerPh {

  Gaussian::Gaussian(int aOrder) {
    fOrder = aOrder;
  } // Gaussian::Gaussian

  Gaussian::~Gaussian() {
    while (fEquations.count() > 0) {
      Vector *equation = fEquations.at(0);
      fEquations.removeFirst();
      delete equation;
    } // while
  } // Gaussian::~Gaussian

  void Gaussian::addToRow(int aRow1, double aFactor, int aRow2) {
    Vector *equation1 = fEquations.at(aRow1);
    Vector *equation2 = fEquations.at(aRow2);
    *equation1 += aFactor * *equation2;
  } // Gaussian::addToRow

  void Gaussian::solve(double aEpsilon) {
    int n = fEquations.count();
    if (n < fOrder) {
      throw std::invalid_argument(qPrintable(QString().sprintf(
        "Gaussian.end(): not enough conditions (%i < %i)", n, fOrder)));
    } // if
    if (n > fOrder) {
      throw std::invalid_argument(qPrintable(QString().sprintf(
        "Gaussian.end(): too many conditions (%i > %i)", n, fOrder)));
    } // if
    /* pivoting: */
    for (int i = 0; i < fOrder; i++) {
      int jMax = 0;
      double pMax = std::fabs(getEquationAt(0) -> getAt(i));
      for (int j = 1; j < fOrder; j++) {
        double p = std::fabs(getEquationAt(j) -> getAt(i));
	if (p > pMax) {
	  pMax = p;
	  jMax = j;
	} // if
      } // for
      if (i != jMax) {
        swapRows(i, jMax);
      } // if
      if (pMax < aEpsilon) {
        throw std::invalid_argument(qPrintable(QString().sprintf(
	  "Gaussian::solve(): not solvable (condition #%i missing)", i)));
      } // if
    } // for
    /* empty lower triangle: */
    for (int i = 0; i < fOrder; i++) {
      multRow(i, 1.0 / getEquationAt(i) -> getAt(i));
      for (int j = i + 1; j < fOrder; j++) {
        addToRow(j, -getEquationAt(j) -> getAt(i), i);
      } // for
    } // for
    /* finally, diagonalize: */
    for (int i = fOrder - 1; i >= 0; i--) {
      for (int j = i - 1; j >= 0; j--) {
        addToRow(j, -getEquationAt(j) -> getAt(i), i);
      } // for
    } // for
  } // Gaussian::solve

  int Gaussian::getCountEquations() {
    return fEquations.count();
  } // Gaussian::getCountEquations

  Vector* Gaussian::getEquationAt(int aRow) {
    return fEquations.at(aRow);
  } // Gaussian::getEquationAt

  int Gaussian::getOrder() {
    return fOrder;
  } // Gaussian::getOrder

  double Gaussian::getValueAt(int aRow) {
    return fEquations.at(aRow) -> getAt(fOrder);
  } // getValueAt

  void Gaussian::multRow(int aRow, double aFactor) {
    Vector *equation = fEquations.at(aRow);
    *equation *= aFactor;
  } // Gaussian::multRow

  Vector* Gaussian::newEquation() {
    Vector *equation = new Vector(fOrder + 1);
    fEquations.append(equation);
    return equation;
  } // Gaussian::newEquation

  void Gaussian::swapRows(int aRow1, int aRow2) {
    if (aRow1 == aRow2) {
      return;
    } // if
    int rowUp = aRow1;
    int rowDown = aRow2;
    if (aRow2 < aRow1) {
      rowUp = aRow2;
      rowDown = aRow1;
    } // if
    Vector *eqDown = fEquations.at(rowDown);
    fEquations.removeAt(rowDown);
    Vector *eqUp = fEquations.at(rowUp);
    fEquations.removeAt(rowUp);
    fEquations.insert(rowUp, eqDown);
    fEquations.insert(rowDown, eqUp);
  } // Gaussian::swapRows

} // GerPh
