/*
  This file is part of GerPhys.

  GerPhys is free software: you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  GerPhys is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  GerPhys.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "projection.h"

#include "floatmap.h"
#include "si.h"
#include <cmath>


namespace GerPh {

  const double Projection::INFINITE_FACTOR = 1.0e9;

  Projection::Projection() {
    fFocalDistance = 1.0;
    fMode = CentralPerspective;
    fInfinity = INFINITE_FACTOR;
    fEpsilon = 1.0 / INFINITE_FACTOR;
    fMap = NULL;
    fDrawing = NULL;
    fDrawArgs = NULL;
    setPhi(0.0);
    setTheta(0.0);
    setPsi(0.0);
  } // Projection::Projection

  Projection::~Projection() {
    if (fMap) {
      FloatMap *map = fMap;
      fMap = NULL;
      map -> setProjection(NULL);
    } // if
  } // Projection::~Projection

  void Projection::cutXYPlane(
    const R3Vector &aRelPos, double &aX, double &aY, bool &aVisible
  ) {
    double f = getFocalDistance();
    double x = aRelPos.getAt(0);
    double y = aRelPos.getAt(1);
    double z = aRelPos.getAt(2);
    aVisible = z < f;
    if (!aVisible) {
      if (std::fabs(x) < fEpsilon) {
        aX = 0.0;
	if (y > 0.0) {
	  aY = fInfinity;
	} else {
	  aY = -fInfinity;
	  return;
	} // if
      } // if
      if (std::fabs(y) < fEpsilon) {
	if (x > 0.0) {
	  aX = fInfinity;
	} else {
	  aX = -fInfinity;
	} // if
        aY = 0.0;
	return;
      } // if
      double ratio = x / y;
      aX = fInfinity;
      aY = ratio * fInfinity;
      return;
    } // if
    double lambda = f / (f - z);
    aX = lambda * x;
    aY = lambda * y;
  } // Projection::cutXYPlane

  void Projection::drawFluff(
    const R3Vector &aCenter, double aRadius, QPainter *aPainter, double aStart
  ) {
    if (!fMap) {
      return;
    } // if
    double x;
    double y;
    double r;
    bool visible;
    getViewSphere(aCenter, aRadius, x, y, r, visible);
    if (!visible) {
      return;
    } // if
    fMap -> drawSpot(x, y, r, aPainter, aStart);
  } // Projection::drawFluff

  void Projection::drawFluffSym (
    const R3Vector &aCenter, int aRadius, QPainter *aPainter, double aStart
  ) {
    if (!fMap) {
      return;
    } // if
    double x;
    double y;
    bool visible;
    getViewXY(aCenter, x, y, visible);
    if (!visible) {
      return;
    } // if
    fMap -> drawSpotSym(x, y, aRadius, aPainter, aStart);
  } // Projection::drawFluffSym

  void Projection::drawLine(
    const R3Vector &aFrom, const R3Vector &aTo, QPainter *aPainter
  ) {
    if (!fMap) {
      return;
    } // if
    double xFrom;
    double yFrom;
    bool visibleFrom;
    double xTo;
    double yTo;
    bool visibleTo;
    getViewXY(aFrom, xFrom, yFrom, visibleFrom);
    getViewXY(aTo, xTo, yTo, visibleTo);
    if (!(visibleFrom || visibleTo)) {
      return;
    } // if
    fMap -> drawLine(xFrom, yFrom, xTo, yTo, aPainter);
  } // Projection::drawLine

  void Projection::drawPoint(const R3Vector &aPosition, QPainter *aPainter) {
    if (!fMap) {
      return;
    } // if
    double x;
    double y;
    bool visible;
    getViewXY(aPosition, x, y, visible);
    if (!visible) {
      return;
    } // if
    fMap -> drawPoint(x, y, aPainter);
  } // Projection::drawPoint

  void Projection::drawSphere(
    const R3Vector &aCenter, double aRadius, QPainter *aPainter
  ) {
    if (!fMap) {
      return;
    } // if
    double x;
    double y;
    double r;
    bool visible;
    getViewSphere(aCenter, aRadius, x, y, r, visible);
    if (!visible) {
      return;
    } // if
    fMap -> drawEllipse(x, y, r, r, aPainter);
  } // Projection::drawSphere

  void Projection::drawSphereSym (
    const R3Vector &aCenter, int aRadius, QPainter *aPainter
  ) {
    if (!fMap) {
      return;
    } // if
    double x;
    double y;
    bool visible;
    getViewXY(aCenter, x, y, visible);
    if (!visible) {
      return;
    } // if
    fMap -> drawEllipseSym(x, y, aRadius, aRadius, aPainter);
  } // Projection::drawSphereSym

  void Projection::drawText(
    const R3Vector &aPosition, const QString &aText, QPainter *aPainter,
    int aHAlign, int aVAlign, int aHOffset, int aVOffset
  ) {
    double x;
    double y;
    bool visible;
    getViewXY(aPosition, x, y, visible);
    if (!visible) {
      return;
    } // if
    fMap -> drawText(x, y, aText, aPainter,
      aHAlign, aVAlign, aHOffset, aVOffset);
  } // Projection::drawText

  void Projection::fillSpace(QBrush *aBrush, QPainter *aPainter) {
    if (!fMap) {
      return;
    } // if
    fMap -> fillAll(aBrush, aPainter);
  } // Projection::fillSpace

  void Projection::getDeviceXY(
    const R3Vector &aPos, int &aX, int &aY, bool &aVisible
  ) {
    if (!fMap) {
      aX = -1;
      aY = -1;
      aVisible = false;
      return;
    } // if
    double x;
    double y;
    getViewXY(aPos, x, y, aVisible);
    aX = fMap -> getXView(x);
    aY = fMap -> getYView(y);
  } // Projection::getDeviceXY

  void* Projection::getDrawArgs() {
    return fDrawArgs;
  } // Projection::getDrawArgs

  void (*Projection::getDrawing())(QPainter*, void*) {
    return fDrawing;
  } // Projection::getDrawing

  double Projection::getFocalDistance() {
    switch(fMode) {
      case CentralPerspective:  return fFocalDistance;
      case ParallelPerspective: return fInfinity;
    } // switch
    return fFocalDistance;
  } // Projection::getFocalDistance

  R3Vector Projection::getFocus() {
    return fPosition + fFocalDistance * getViewDirection();
  } // Projection::getFocus

  FloatMap* Projection::getMap() {
    return fMap;
  } // Projection::getMap

  Projection::Mode Projection::getMode() {
    return fMode;
  } // Projection::getMode

  double Projection::getPhi() {
    return fPhi;
  } // Projection::getPhi

  double Projection::getPhi_deg() {
    return fPhi / SI::unit_deg;
  } // Projection::getPhi_deg

  GerPh::R3Vector Projection::getPosition() {
    return fPosition;
  } // Projection::getPosition

  double Projection::getPsi() {
    return fPsi;
  } // Projection::getPsi

  double Projection::getPsi_deg() {
    return fPsi / SI::unit_deg;
  } // Projection::getPsi_deg

  R3Matrix Projection::getRotation() {
    if (!fRotValid) {
      fRotation = fRotPsi * fRotTheta * fRotPhi;
    } // if
    return fRotation;
  } // Projection::getRotation

  double Projection::getTheta() {
    return fTheta;
  } // Projection::getTheta

  double Projection::getTheta_deg() {
    return fTheta / SI::unit_deg;
  } // Projection::getTheta_deg

  R3Vector Projection::getViewDirection() {
    double cos_theta = std::cos(fTheta);
    double sin_theta = std::sin(fTheta);
    double cos_phi = std::cos(fPhi);
    double sin_phi = std::sin(fPhi);
    return R3Vector(
      sin_theta * sin_phi,
      -sin_theta * cos_phi,
      cos_theta
    );
  } // Projection::getViewDirection

  void Projection::getViewSphere(
    const R3Vector &aCenter, double aRadius,
    double &aX, double &aY, double &aR, bool &aVisible
  ) {
    switch(fMode) {
      case CentralPerspective:
        getViewSphereCentral(aCenter, aRadius, aX, aY, aR, aVisible); break;
      case ParallelPerspective:
        getViewSphereParallel(aCenter, aRadius, aX, aY, aR, aVisible); break;
    } // switch
  } // Projection::getViewSphere

  void Projection::getViewSphereCentral(
    const R3Vector &aCenter, double aRadius,
    double &aX, double &aY, double &aR, bool &aVisible
  ) {
    aX = 0.0;
    aY = 0.0;
    aR = 0.0;
    aVisible = false;
    R3Vector relCenter = getRotation() * (aCenter - fPosition);
    double f = getFocalDistance();
    double z = relCenter.getAt(2) - f;
    if (z > 0.0) {
      return;
    } // if
    double x = relCenter.getAt(0);
    double y = relCenter.getAt(1);
    R3Vector diffCenter;
    diffCenter.setXYZ(x, y, z);
    double distCenter = diffCenter.getAbs();
    if (aRadius > distCenter) {
      return;
    } // if
    double cosDelta = -z / distCenter;
    double delta = std::acos(cosDelta);
    /* Put sphere on z-axis, calculate tangent lines: */
    R3Vector tangent1;
    R3Vector tangent2;
    double R_a = aRadius / distCenter;
    double sqrt_aR = sqrt(distCenter * distCenter - aRadius * aRadius);
    double tx = R_a * aRadius;
    double tz = -R_a * sqrt_aR;
    tangent1.setXYZ(+tx, 0.0, tz);
    tangent2.setXYZ(-tx, 0.0, tz);
    /*
      Move sphere back to its actual position, first with a rotation around the
      y-axis ...
    */
    R3Matrix rotation;
    rotation.beRotation(delta, 1);
    tangent1 = rotation * tangent1;
    tangent2 = rotation * tangent2;
    /*
      ... then with a rotation around the z-axis ...
    */
    double chi = SI::CONST_pi * 0.5;
    if (std::fabs(x) > fEpsilon) {
      chi = std::atan(y / x);
      if (x > fEpsilon) {
        chi += SI::CONST_pi;
      } // if
    } else if (y > 0.0) {
      chi = -chi;
    } // if
    rotation.beRotation(chi, 2);
    tangent1 = rotation * tangent1;
    tangent2 = rotation * tangent2;
    /*
      ... then with a translation along the focus vector:
    */
    tangent1.setAt(2, tangent1.getAt(2) + f);
    tangent2.setAt(2, tangent2.getAt(2) + f);
    /*
      Finally, calculate the projection of the tangent lines to the X,Y-plane.
      Use their average as center of the projected circle, their (half)
      difference as its radius:
    */
    double x1;
    double y1;
    bool visible1;
    double x2;
    double y2;
    bool visible2;
    cutXYPlane(tangent1, x1, y1, visible1);
    cutXYPlane(tangent2, x2, y2, visible2);
    aX = 0.5 * (x1 + x2);
    aY = 0.5 * (y1 + y2);
    double dx = x1 - x2;
    double dy = y1 - y2;
    aR = 0.5 * sqrt(dx * dx + dy * dy);
    aVisible = visible1 || visible2;
  } // Projection::getViewSphereCentral

  void Projection::getViewSphereParallel(
    const R3Vector &aCenter, double aRadius,
    double &aX, double &aY, double &aR, bool &aVisible
  ) {
    aX = 0.0;
    aY = 0.0;
    aR = aRadius;
    R3Vector focalVector(fInfinity, 0.0, 0.0);
    R3Vector diffCenter = aCenter - focalVector;
    aVisible =
      (diffCenter.getAbs() > aRadius) &&
      (aCenter.getAt(2) - aRadius < fInfinity);
    bool irrelevant;
    getViewXY(aCenter, aX, aY, irrelevant);
  } // Projection::getViewSphereParallel

  void Projection::getViewXY(
    const R3Vector &aPosition, double &aX, double &aY, bool &aVisible
  ) {
    cutXYPlane(getRotation() * (aPosition - fPosition), aX, aY, aVisible);
  } // Projection::getViewXY

  void Projection::rescale() {
    if (!fMap) {
      fInfinity = INFINITE_FACTOR;
      fEpsilon = 1.0 / INFINITE_FACTOR;
      return;
    } // if
    double d = fMap -> getDiagonal();
    fInfinity = d * INFINITE_FACTOR;
    fEpsilon = d / INFINITE_FACTOR;
  } // Projection::rescale

  void Projection::setDrawArgs(void *aValue) {
    fDrawArgs = aValue;
  } // Projection::setDrawArgs

  void Projection::setDrawing(void (*aValue)(QPainter*, void*)) {
    fDrawing = aValue;
  } // Projection::setDrawing

  void Projection::setFocalDistance(double aValue) {
    if (std::isnan(1.0 / aValue)) {
      return; //!!! TODO: throw an exception?
    } // if
    fFocalDistance = aValue;
  } // Projection::setFocalDistance

  void Projection::setMap(FloatMap *aValue) {
    if (fMap == aValue) {
      return;
    } // if
    if (fMap) {
      fMap -> setProjection(NULL);
    } // if
    fMap = aValue;
    if (fMap) {
      fMap -> setProjection(this);
    } // if
  } // Projection::setMap

  void Projection::setMode(Projection::Mode aValue) {
    fMode = aValue;
  } // Projection::setMode

  void Projection::setPhi(double aValue) {
    fPhi = aValue;
    fRotPhi.beRotation(-aValue, 2);
    fRotValid = false;
  } // Projection::setPhi

  void Projection::setPhi_deg(double aValue) {
    setPhi(aValue * SI::unit_deg);
  } // Projection::setPhi_deg

  void Projection::setPosition(const R3Vector &aValue) {
    fPosition = aValue;
  } // Projection::setPosition

  void Projection::setPsi(double aValue) {
    fPsi = aValue;
    fRotPsi.beRotation(-aValue, 2);
    fRotValid = false;
  } // Projection::setPsi

  void Projection::setPsi_deg(double aValue) {
    setPsi(aValue * SI::unit_deg);
  } // Projection::setPsi_deg

  void Projection::setTheta(double aValue) {
    fTheta = aValue;
    fRotTheta.beRotation(-aValue, 0);
    fRotValid = false;
  } // Projection::setTheta

  void Projection::setTheta_deg(double aValue) {
    setTheta(aValue * SI::unit_deg);
  } // Projection::setTheta_deg

} // GerPh
