/*
  This file is part of GerPhys.

  GerPhys is free software: you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  GerPhys is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  GerPhys.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef GERPH_PROJECTION_H
#define GERPH_PROJECTION_H


#include "floatmap.h"
#include "gerphysbase.h"
#include "r3matrix.h"
#include "r3vector.h"
#include <QHash>


namespace GerPh {


  class Projection {

    public:

      Projection();
      virtual ~Projection();

      double getPhi();
      void setPhi(double aValue);
      double getPhi_deg();
      void setPhi_deg(double aValue);
      double getTheta();
      void setTheta(double aValue);
      double getTheta_deg();
      void setTheta_deg(double aValue);
      double getPsi();
      void setPsi(double aValue);
      double getPsi_deg();
      void setPsi_deg(double aValue);

      R3Vector getPosition();
      void setPosition(const R3Vector &aValue);

      R3Vector getViewDirection();

      R3Vector getFocus();

      typedef enum {
        CentralPerspective,
	ParallelPerspective
      } Mode;

      Mode getMode();
      void setMode(Mode aValue);

      double getFocalDistance();
      void setFocalDistance(double aValue);

      void getViewXY(const R3Vector &aPosition, double &aX, double &aY,
        bool &aVisible);
      void getViewSphere(const R3Vector &aCenter, double aRadius,
        double &aX, double &aY, double &aR, bool &aVisible);

      void getDeviceXY(const R3Vector &aPosition, int &aX, int &aY,
        bool &aVisible);

      void drawPoint(const R3Vector &aPosition, QPainter *aPainter);
      void drawLine(const R3Vector &aFrom, const R3Vector &aTo,
        QPainter *aPainter);
      void drawSphere(const R3Vector &aCenter, double aRadius,
        QPainter *aPainter);
      void drawSphereSym(const R3Vector &aCenter, int aRadius,
        QPainter *aPainter);
      void drawFluff(const R3Vector &aCenter, double aRadius,
        QPainter *aPainter, double aStart = 0.0);
      void drawFluffSym(const R3Vector &aCenter, int aRadius,
        QPainter *aPainter, double aStart = 0.0);
      void drawText(const R3Vector &aPosition, const QString &aText,
        QPainter *aPainter,
	int aHAlign = FloatMap::ALIGN_CENTER,
	int aVAlign = FloatMap::ALIGN_CENTER,
	int aHOffset = 0, int aVOffset = 0);

      void fillSpace(QBrush *aBrush, QPainter *aPainter);

      FloatMap *getMap();
      void setMap(FloatMap *aValue);

      void (*getDrawing())(QPainter*, void*);
      void setDrawing(void (*aValue)(QPainter*, void*));
      void *getDrawArgs();
      void setDrawArgs(void *aValue);

      void rescale();

    private:

      double fPhi;
      double fTheta;
      double fPsi;

      R3Vector fPosition;

      R3Matrix fRotPhi;
      R3Matrix fRotTheta;
      R3Matrix fRotPsi;
      R3Matrix fRotation;
      bool fRotValid;

      R3Matrix getRotation();

      double fFocalDistance;
      Mode fMode;

      FloatMap *fMap;

      void (*fDrawing)(QPainter*, void*);
      void *fDrawArgs;

      double fInfinity;
      double fEpsilon;

      void cutXYPlane(const R3Vector &aRelPos, double &aX, double &aY,
        bool &aVisible);
      void getViewSphereCentral(const R3Vector &aCenter, double aRadius,
        double &aX, double &aY, double &aR, bool &aVisible);
      void getViewSphereParallel(const R3Vector &aCenter, double aRadius,
        double &aX, double &aY, double &aR, bool &aVisible);

      static const double INFINITE_FACTOR;

  }; // Projection


} // GerPh


#endif
